/* -*- Mode: C -*- */

// ===========================================================================
//
//   P O W E R    S U P P L Y    C O N T R O L L E R
//
//   This file:   C O N T R O L L E R . H
//
//   $Id: controller.h,v 1.5 2009/01/06 23:13:08 frgo Exp frgo $
//
// ===========================================================================

#if !defined( __CONTROLLER_H__ )
#define __CONTROLLER_H__

#ifdef __cplusplus
extern "C" {
#endif

  // --------------------------------------------------------------------------
  // DEFINITIONS 
  // --------------------------------------------------------------------------

#define PSU_MODEL      "P1025A"
#define PSU_VERSION    "X.01.00"
#define PSU_SERIAL_NR  "801"
 
#define PSU_BUFFER_LENGTH            LCD_LINE_LENGTH

#define HEART_BEAT_TICKS             5000
#define LAMP_BLINK_TICKS             3333

  // DEBUG STUFF

#define PSU_DEBUG_LEVEL_NONE            0
#define PSU_DEBUG_LEVEL_VERBOSE_1    1000
#define PSU_DEBUG_LEVEL_VERBOSE_2    2000    

  // RETURN CODES: INTERNAL "COMMANDS"

#define RC_COMMANDS_BASE        0

#define ISR_DO_SHUTDOWN         (RC_COMMANDS_BASE +    1)
#define ISR_DO_RESET            (RC_COMMANDS_BASE +    2)
#define ISR_DO_FAIL             (RC_COMMANDS_BASE +    3) 

#define RC_ALREADY_LOCKED       (RC_COMMANDS_BASE +    9)

#define NO_CMD                        0

  // RETURN CODES: ERRORS

  // DARC ERRORS - Start at +100

#define RC_ERROR_INVALID_DARC_COMMAND          (ERROR_RECOVERABLE_BASE + 101) 
#define RC_ERROR_OPERATION_NOT_PERMITTED       (ERROR_RECOVERABLE_BASE + 102)
#define RC_ERROR_FUNCTION_NOT_IMPLEMENTED      (ERROR_RECOVERABLE_BASE + 103)

  // AVR C ERRORS

#define RC_ERROR_WATCHDOG_FAILED               (ERROR_INTERNAL_BASE    + 601) 

  // RELAY / FAN / VOLTAGE / IO ERRORS

#define RC_ERROR_INVALID_RELAY_STATE           (ERROR_INTERNAL_BASE    + 701)
#define RC_ERROR_INVALID_VOLTAGE_COMBINATION   (ERROR_INTERNAL_BASE    + 702)
#define RC_ERROR_INVALID_FAN_SPEED             (ERROR_INTERNAL_BASE    + 703)
#define RC_ERROR_INVALID_IO_ID                 (ERROR_INTERNAL_BASE    + 704)
#define RC_ERROR_STREAM_ERROR                  (ERROR_INTERNAL_BASE    + 705)
#define RC_ERROR_VOLTAGE_NOT_FOUND             (ERROR_INTERNAL_BASE    + 706)
#define RC_ERROR_RELAY_NOT_FOUND               (ERROR_INTERNAL_BASE    + 707)
#define RC_ERROR_LAMP_NOT_FOUND                (ERROR_INTERNAL_BASE    + 708)
#define RC_ERROR_FAN_NOT_FOUND                 (ERROR_INTERNAL_BASE    + 709)

  // I2C APPLICATION ERRORS

#define RC_ERROR_INVALID_I2C_ADDRESS           (ERROR_INTERNAL_BASE    + 801) 

  // CONTROLLER SW ERRORS

#define RC_ERROR_ILLEGAL_STATUS_ID             (ERROR_INTERNAL_BASE    + 901)
#define RC_ERROR_ILLEGAL_STATUS_CHANGE         (ERROR_INTERNAL_BASE    + 902)

  // STATUS DEFINITIONS

#define DISPLAY_STATUS_LINE           4 

  // PSU STATUSES

#define PSU_STATUS_ID_PSU_OFF             0
#define PSU_STATUS_TEXT_PSU_OFF           "OFF"

#define PSU_STATUS_ID_PSU_BOOTING         1
#define PSU_STATUS_TEXT_PSU_BOOTING       "BOOT"

#define PSU_STATUS_ID_PSU_STANDBY         2
#define PSU_STATUS_TEXT_PSU_STANDBY       "STDBY"

#define PSU_STATUS_ID_PSU_OPERATING       3
#define PSU_STATUS_TEXT_PSU_OPERATING     "OPER"

#define PSU_STATUS_ID_PSU_IN_TEST_MODE    4
#define PSU_STATUS_TEXT_PSU_IN_TEST_MODE  "TEST"

#define PSU_STATUS_ID_PSU_FAILURE         5
#define PSU_STATUS_TEXT_PSU_FAILURE       "FAIL"

#define PSU_STATUS_ID_PSU_SHUTTING_DOWN   6
#define PSU_STATUS_TEXT_PSU_SHUTTING_DOWN "SHUT"

#define PSU_STATUS_ID_PSU_RESETTING       7
#define PSU_STATUS_TEXT_PSU_RESETTING     "RESET"

#define PSU_STATUS_ID_PSU_POWER_ON        8
#define PSU_STATUS_TEXT_PSU_POWER_ON      "PWR ON"

#define PSU_STATUS_TEXT_NA                "N/A"

#define PSU_STATUS_ID_NO_STATUS           254

  typedef uint8_t tStatusID;

  typedef struct _tsStatusDefinition
  {
    tStatusID nID;
    char *pcText;
    
  } tsStatusDefinition, * tpsStatusDefinitionPtr;

#define STATUS_CHANGED() (nCurrentStatus != nNextStatus)
#define BREAK_IF_STATUS_CHANGED() if( STATUS_CHANGED() ) { break; }

  // FAILURE REASON CODES

#define PSU_FAILURE_REASON_EXTERNAL_FAILURE     1010

#define PSU_FAILURE_REASON_MAINS_VOLTAGE_LOW    2010


  // I2C I/O Cards

#define I2C_CARD_0                       0

#define NO_I2C                         255

#define I2C_CARD_0_OUTPUT_ADDRESS       64
#define I2C_CARD_0_INPUT_ADDRESS       112

#define IS_VALID_I2C_ADDRESS(address) ((address > 0) && (address <= 128))

  typedef struct _tsI2CCard
  {
    uint8_t nID;
    uint16_t nOutAddress;
    uint16_t nInAddress;

  } tsI2CCard, *tpsI2CCardPtr;

#define IO_OFF  0
#define IO_ON   1

  // OBJECT CLASS IDS

#define OBJECT_CLASS_ID_RELAY                   0
#define OBJECT_CLASS_ID_VOLTAGE                 1
#define OBJECT_CLASS_ID_FAN                     2
#define OBJECT_CLASS_ID_LAMP                    3

#define OBJECT_CLASS_ID_PSU_CONFIGURATION     251
#define OBJECT_CLASS_ID_FAILURE_REASONCODE    252
#define OBJECT_CLASS_ID_STATUS                253
#define OBJECT_CLASS_ID_ERRORLOG              254

  // RELAYS

#define RELAY_ID_UA                      0
#define RELAY_ID_UG2                     1
#define RELAY_ID_UG1                     2
#define RELAY_ID_UH                      3
#define RELAY_ID_HV_ENABLE               4
#define RELAY_ID_MAIN_POWER              5
#define RELAY_ID_MAIN_FAN_SPEED          6

#define NO_RELAY                         255

#define RELAY_OFF                        IO_OFF
#define RELAY_ON                         IO_ON

  typedef struct _tsRelay
  {
    uint8_t  nID;
    uint8_t  nI2CCardID;
    uint8_t  nState;
    uint16_t nLampID;

  } tsRelay, *tpsRelayPtr;

  // Voltages by Name (and translations to the relevant Relay ID)

#define VOLTAGE_ID_UA           RELAY_ID_UA
#define VOLTAGE_ID_UG2          RELAY_ID_UG2
#define VOLTAGE_ID_UG1          RELAY_ID_UG1
#define VOLTAGE_ID_UH           RELAY_ID_UH

#define NO_VOLTAGE              NO_RELAY

#define VOLTAGE_OFF             RELAY_OFF
#define VOLTAGE_ON              RELAY_ON

#define VOLTAGE_ADC_MUX_0       (1<<0)
#define VOLTAGE_ADC_MUX_1       (1<<1)
#define VOLTAGE_ADC_MUX_2       (1<<2)
#define VOLTAGE_ADC_MUX_3       (1<<3)
       
#define ADC_PRESCALE            (F_CPU / 125000)

#define VOLTAGE_RANGE_NOT_INITIALIZED_VALUE 0.0

  typedef struct _tsVoltage
  {
    uint8_t nID;
    uint8_t nADCMux;
    uint8_t nState;
    double  fRangeMin;
    double  fRangeMax;
    double  nValue;

  } tsVoltage, *tpsVoltagePtr;

  // Fans

#define FAN_ID_MAIN_FAN         0

#define FAN_OFF                 0
#define FAN_SPEED_1             50
#define FAN_SPEED_2             100
#define FAN_SPEED_MAX           FAN_SPEED_2

#define PSU_SHUTDOWN_COOLDOWN_TIME   10 // [s]
#define PSU_FAIL_COOLDOWN_TICKS    1000 // Ticks (not seconds !)

  typedef struct _tsFan
  {
    uint8_t nID;
    uint8_t nRelayID;
    uint8_t nState;

  } tsFan, *tpsFanPtr;
 
  // Frontpanel push buttons

#define PSU_FP_SW_MAIN_POWER    (1<<0)

#define PSU_FP_PB_OPER          (1<<1)
#define PSU_FP_PB_TEST          (1<<2)

#define PSU_FP_PB_HV_ENABLE     (1<<3)
#define PSU_FP_PB_UA            (1<<4)
#define PSU_FP_PB_UG2           (1<<5)
#define PSU_FP_PB_UG1           (1<<6)
#define PSU_FP_PB_UH            (1<<7)

#define PSU_FP_PB_RESET         (1<<8)

#define PSU_ALL_KEYS_MASK       ( PSU_FP_SW_MAIN_POWER | \
                                  PSU_FP_PB_OPER |       \
                                  PSU_FP_PB_TEST |       \
                                  PSU_FP_PB_HV_ENABLE |  \
                                  PSU_FP_PB_UA |         \
                                  PSU_FP_PB_UG2 |        \
                                  PSU_FP_PB_UG1 |        \
                                  PSU_FP_PB_UH |         \
                                  PSU_FP_PB_RESET )
                                 


#define I2C_ADDRESS_FP_PB          I2C_CARD_0_INPUT_ADDRESS

#define KEY_PRESSED(pbs,keybit)    (pbs & keybit)


#define POWER_SWITCH_ON(pbs)       KEY_PRESSED(pbs,PSU_FP_SW_MAIN_POWER) 

#define OPER_PB_PRESSED(pbs)       KEY_PRESSED(pbs,PSU_FP_PB_OPER)
#define TEST_PB_PRESSED(pbs)       KEY_PRESSED(pbs,PSU_FP_PB_TEST)

#define HV_ENABLE_PB_PRESSED(pbs)  KEY_PRESSED(pbs,PSU_FP_PB_HV_ENABLE)
#define UA_PB_PRESSED(pbs)         KEY_PRESSED(pbs,PSU_FP_PB_UA)
#define UG2_PB_PRESSED(pbs)        KEY_PRESSED(pbs,PSU_FP_PB_UG2)
#define UG1_PB_PRESSED(pbs)        KEY_PRESSED(pbs,PSU_FP_PB_UG1)
#define UH_PB_PRESSED(pbs)         KEY_PRESSED(pbs,PSU_FP_PB_UH)

#define RESET_PB_PRESSED(pbs)      KEY_PRESSED(pbs,PSU_FP_PB_RESET)

  typedef struct _tsPB
  {
    uint16_t nID;
    uint8_t  nI2CCardID;
    uint16_t nLampID;
    uint8_t  nRelayID;

  } tsPB, *tpsPBPtr;

  // Push button definitions

#define PSU_FP_PB_READ_INTERVAL       0.1  // [s]
#define PSU_IO_TIMER_PRESCALER_VALUE  1

  // Frontpanel lamps

#define LAMP_ID_MAIN_POWER        0

#define LAMP_ID_OPER              1
#define LAMP_ID_TEST              2

#define LAMP_ID_HV_ENABLE         3
#define LAMP_ID_UA                4
#define LAMP_ID_UG2               5
#define LAMP_ID_UG1               6
#define LAMP_ID_UH                7

#define LAMP_ID_RESET             8
#define LAMP_ID_FAIL              9

#define NO_LAMP                 255

#define LAMP_STATE_OFF          IO_OFF
#define LAMP_STATE_ON           IO_ON
#define LAMP_STATE_BLINK        (IO_ON + IO_OFF + 2)
#define LAMP_BLINK              TRUE
#define LAMP_DONT_BLINK         FALSE

  typedef struct _tsLamp
  {
    uint16_t nID;
    uint8_t  nI2CCardID;
    uint8_t  nState;
    bool_t   bBlink;
  } tsLamp, *tpsLampPtr;

  // DARC (= CAT) COMMANDS

  // Complete CAT command string = 
  // PREFIX + SERIAL_NO + COMAND TEXT
  // Each part to be separate by a "-" !

  // Control of voltages
  // All SWITCH commands have a parameter being either "ON" or "OFF
  // sepatared by a blank.
  // The pure voltage commands get or set the voltage value.
  // The voltage always has 4 significant digits:
  // UA: 0 .. 5000 V => "0000" -> "0345" -> "5000" are valid value examples
  // UH: 0 .. 30.00 V => "00.00" -> "01.34" -> "30.00"
  // UG1: 0 .. -200 V => "000.0" -> "-200.0"
  // UG2: 0 .. 500 V => "000.0" -> "500.0"

  // SWITCHING:
  // "SWITCH UA ON", "SWITCH UH OFF" etc.

#define DARC_CAT_CMD_ID_SWITCH_VOLTAGE_OFF     1
#define DARC_CAT_CMD_ID_SWITCH_VOLTAGE_ON      2

#define DARC_CAT_CMD_ID_GET_VOLTAGE            3
#define DARC_CAT_CMD_ID_SET_VOLTAGE            4

#define DARC_CAT_CMD_ID_PSU_GET_STATE          5

#define DARC_CAT_CMD_ID_ENABLE_HV              6
#define DARC_CAT_CMD_ID_DISABLE_HV             7
#define DARC_CAT_CMD_ID_GET_HV_STATE           8

#define DARC_CAT_CMD_ID_PSU_RESET              9
#define DARC_CAT_CMD_ID_PSU_SHUTDOWN          10

#define DARC_CAT_CMD_SWITCH_VOLTAGE          "SWITCH-VOLTAGE"
#define DARC_CAT_CMD_VOLTAGE                 "VOLTAGE"
#define DARC_CAT_CMD_UA                      "UA"
#define DARC_CAT_CMD_UG2                     "UG2"
#define DARC_CAT_CMD_UG1                     "UG1"
#define DARC_CAT_CMD_UH                      "UH"
#define DARC_CAT_CMD_PSU                     "PSU"

  // STATE REQUEST / RESPONSE
  // "UG-STATE?" -> "UH-STATE OFF" or "UH-STATE ON" etc.
  
#define DARC_CAT_CMD_STATE             "STATE"

#define DARC_CAT_CMD_ENABLE_HV         "ENABLE-HV"    // -> "OK" | "FAIL"
#define DARC_CAT_CMD_DISABLE_HV        "DISABLE-HV"   // -> "OK" | "FAIL"
#define DARC_CAT_CMD_HV_ENABLED        "HV-ENABLED?"  // -> "YES" "NO"

#define DARC_CAT_CMD_PSU_RESET         "PSU-RESET"
#define DARC_CAT_CMD_PSU_STATE         "PSU-STATE"
#define DARC_CAT_CMD_PSU_SHUTDOWN      "PSU-SHUTDOWN"

#define DARC_CAT_CMD_GO_STANDBY        "STANDBY"


  typedef struct _tsVoltageInfo
  {
    uint32_t nVoltageID;
    double   fValue;
    uint16_t nState;

  } tsVoltageInfo, *tpsVoltageInfoPtr;

  // Events

#define PSU_EVENT_ID_STATE_CHANGE     1
#define PSU_EVENT_ID_BOOT             2
#define PSU_EVENT_ID_FAIL             4

  // DARC Request/Response Data Structure

  typedef union
  {
    tsVoltageInfo sVoltageInfo;
    uint8_t       nHVState;
    uint16_t      nPSUStatusID;
       
  } tuDARCData;

  // Function prototypes

  // General utility functions

  uint16_t    lock( void );
  void        unlock( void );

  void        enable_i2c( void );
  void        disable_i2c( void );

  void        set_debug_level( uint16_t );

  // General I/O functions

  void        init_io_timer( void );

  // Digital Appliance Remote Control (DARC) functions

  rc_t        process_darc( void );
  rc_t        process_darc_command( tpsDARCCommandPtr );

  rc_t        darc_send_psu_status( uint8_t );
  rc_t        darc_send_errorlog( void );
  rc_t        darc_send_relay( uint8_t );
  rc_t        darc_send_voltage( uint8_t );
  rc_t        darc_send_fan( uint8_t );
  rc_t        darc_send_lamp( uint8_t );
  rc_t        darc_send_failure_reasoncode( void );
  rc_t        darc_send_psu_configuration( void );

  // Status related functions

  tStatusID   get_status( void );
  void        set_status( tStatusID );
  char *      get_status_text( tStatusID );

  void        update_status_display( void );
  rc_t        switch_status( tStatusID );
  rc_t        goto_next_status( tStatusID );

  void        heart_beat( void );

  rc_t        print_status( FILE * );

  // I2C I/O control

  tpsI2CCardPtr get_i2ccard( uint16_t );
  rc_t          switch_io( uint8_t, uint8_t, uint8_t );
 
  // Pushbutton control

  uint16_t    get_pb_sample( void );
  void        debounce( void );
  
  uint16_t    get_pbs( uint16_t );
  uint16_t    wait_for_pbs( uint16_t );
  uint16_t    wait_for_valid_pb( void );
  void        wait_for_power_on( void );
  void        wait_for_power_off( void );
  rc_t        set_accepted_pbs( uint16_t );

  // Relay control functions
 
  uint16_t    relay_id_to_i2c_address( uint16_t );
  tpsRelayPtr get_relay( uint16_t );
  rc_t        switch_relay( uint16_t, uint8_t );
  rc_t        init_relays( void );
  rc_t        toggle_relay( uint16_t );
  
  rc_t        print_relay( uint16_t, FILE * );
  rc_t        print_relays( FILE * );
  
  // Voltage control functions

  rc_t        switch_voltage( uint16_t, uint8_t );
  rc_t        switch_off_all_voltages( void );
  rc_t        toggle_voltage( uint16_t );
  rc_t        bring_up_voltages( void );
  rc_t        bring_down_voltages( void );
  rc_t        check_voltage_combination( uint16_t, uint8_t );
  void        set_check_voltage_combination_flag( uint8_t );

  rc_t        print_voltage( uint16_t, FILE * );
  rc_t        print_voltages( FILE * );

  // Fan control functions

  rc_t        switch_fan( uint8_t, uint8_t );
  
  rc_t        print_fan( uint8_t, FILE * );
  rc_t        print_fans( FILE * );

  // Frontpanel lamp control
 
  tpsLampPtr  get_lamp( uint16_t );
  rc_t        set_lamp_status( uint16_t, uint8_t );
  rc_t        switch_lamp( uint16_t, uint8_t );
  rc_t        update_lamps( void );
  rc_t        update_lamps_on_frontpanel( void );
  rc_t        blink_lamps( void );

  rc_t        print_lamp( uint16_t, FILE * );
  rc_t        print_lamps( FILE * );

  // Controller init functions

  rc_t        init_0( void );
  rc_t        init_1( void );

  rc_t        greet( void );

  // PSU helper functions
 
  void        analyze_reset_reason( void );
     
  rc_t        print_object( uint8_t, uint8_t, FILE * ); 

  rc_t        print_failure_reasoncode( FILE * );
  rc_t        set_failure_reasoncode( uint16_t );
  uint16_t    get_failure_reasoncode( void );

  rc_t        print_psu_configuration( FILE * );

  void        execute_cmd_from_isr( void );

  // Main PSU functions

  rc_t        psu_off( void );
  rc_t        boot( void );
  rc_t        standby( void );
  rc_t        operate( void );
  rc_t        test_mode( void );
  rc_t        fail( uint16_t );
  rc_t        shutdown( void );
 
  int         main( void );
 
#ifdef __cplusplus
} /* closing brace for extern "C" */
#endif

#endif
